from strictyaml import load, Map, Str, Seq, Optional, MapPattern, Any, YAML
from hitchstory.result import ResultList, Success, Failure
from hitchstory.arguments import Arguments
from hitchstory import exceptions
from hitchstory import utils
from slugify import slugify
from path import Path
import prettystack
import strictyaml
import inspect
import time
import copy


THIS_DIRECTORY = Path(__file__).realpath().dirname()


DEFAULT_STACK_TRACE = prettystack.PrettyStackTemplate()\
                                 .to_console()\
                                 .cut_calling_code(
                                    THIS_DIRECTORY.joinpath("story.py"))\


class StoryStep(object):
    def __init__(self, story, yaml_step, index, params):
        self._yaml = yaml_step
        self._story = story
        self._index = index
        if isinstance(yaml_step.value, str):
            self.name = str(yaml_step)
            self.arguments = Arguments(None, params)
        elif isinstance(yaml_step.value, dict) and len(yaml_step.keys()) == 1:
            self.name = list(yaml_step.keys())[0]
            self.arguments = Arguments(list(yaml_step.values())[0], params)
        else:
            raise RuntimeError("Invalid YAML in step '{}'".format(yaml_step))

    def underscore_case_name(self):
        return utils.to_underscore_style(str(self.name))

    def update(self, **kwargs):
        self._story.update(self, kwargs)

    @property
    def index(self):
        return self._index

    @property
    def yaml(self):
        return self._yaml

    def run(self, engine):
        if hasattr(engine, self.underscore_case_name()):
            attr = getattr(engine, self.underscore_case_name())
            if hasattr(attr, '__call__'):
                step_method = attr

                validators = step_method._validators \
                    if hasattr(step_method, '_validators') else {}
                self.arguments.validate(validators)

                if self.arguments.is_none:
                    step_method()
                elif self.arguments.single_argument:
                    if type(self.arguments.argument) is str:
                        step_method(self.arguments.argument)
                    else:
                        step_method(self.arguments.argument.value)
                else:
                    argspec = inspect.getargspec(step_method)

                    if argspec.keywords is not None:
                        kwargs = {
                            key.data: val for key, val in \
                                self.arguments.kwargs.items()
                        }
                        step_method(**kwargs)
                    else:
                        step_method(**self.arguments.pythonized_kwargs())
            else:
                raise exceptions.StepNotCallable((
                    "Step with name '{}' in {} is not a function "
                    "or a callable object, it is a {}".format(
                        self.underscore_case_name(),
                        engine.__repr__(),
                        type(attr)
                    )
                ))
        else:
            raise exceptions.StepNotFound("Step with name '{}' not found in {}.".format(
                self.underscore_case_name(),
                engine.__repr__()
            ))


class Story(object):
    def __init__(self, story_file, name, parsed_yaml, engine, collection):
        self._story_file = story_file
        self._name = name
        self._parsed_yaml = parsed_yaml
        self._engine = engine
        self._steps = []
        self._about = {}
        if engine.schema.about is not None:
            for about_property in engine.schema.about.keys():
                self._about[about_property] = parsed_yaml.get(about_property)
        self._collection = collection

    def update(self, step, kwargs):
        self._story_file.update(self, step, kwargs)

    @property
    def story_file(self):
        return self._story_file

    @property
    def filename(self):
        return self._story_file.filename

    @property
    def about(self):
        return self._about

    @property
    def name(self):
        return self._name

    @property
    def slug(self):
        return slugify(self._name)

    @property
    def params(self):
        param_dict = self._collection.named(str(self._parsed_yaml['based on'])).params \
            if "based on" in self._parsed_yaml else {}
        for name, param in self._parsed_yaml.get("params", {}).items():
            param_dict[name] = param
        return param_dict

    def unparameterized_preconditions(self):
        precondition_dict = {}
        precondition_dict = self._collection.named(
            str(self._parsed_yaml['based on'])
        ).unparameterized_preconditions() \
            if "based on" in self._parsed_yaml else {}
        for name, precondition in self._parsed_yaml.get("preconditions", {}).items():
            precondition_dict[str(name)] = precondition
        return precondition_dict

    @property
    def preconditions(self):
        precondition_dict = self.unparameterized_preconditions()
        for name, precondition in precondition_dict.items():
            precondition_value = precondition.data \
                if isinstance(precondition, YAML) else precondition
            for param_name, param in self.params.items():
                precondition_value = utils.replace_parameter(precondition_value, param_name, param)
            precondition_dict[str(name)] = precondition_value
        return precondition_dict

    @property
    def steps(self):
        step_list = self._collection.named(str(self._parsed_yaml['based on'])).steps \
            if "based on" in self._parsed_yaml else []
        step_list.extend(self._parsed_yaml.get('scenario', []))
        return step_list

    @property
    def scenario(self):
        return [
            StoryStep(
                self, parsed_step, index, self.params
            ) for index, parsed_step in enumerate(self.steps)
        ]

    def run_special_method(self, method, exception_to_raise):
        try:
            method()
        except Exception as exception:
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exception_to_raise(
                stack_trace
            )

    def play(self):
        """
        Run a story from beginning to end, time it and return
        a Result object.
        """
        start_time = time.time()
        passed = False
        caught_exception = None
        try:
            from signal import signal, SIGINT, SIGTERM, SIGHUP, SIGQUIT

            signal(SIGINT, self._engine.on_abort)
            signal(SIGTERM, self._engine.on_abort)
            signal(SIGHUP, self._engine.on_abort)
            signal(SIGQUIT, self._engine.on_abort)

            current_step = None
            self._engine._preconditions = self.preconditions
            self._engine.story = self
            self._engine.set_up()

            for step in self.scenario:
                current_step = step
                self._engine.current_step = current_step
                step.run(self._engine)
            passed = True
        except Exception as exception:
            caught_exception = exception
            failure_stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

        if passed:
            self.run_special_method(self._engine.on_success, exceptions.OnSuccessException)
            result = Success(self, time.time() - start_time)
        else:
            self.run_special_method(self._engine.on_failure, exceptions.OnFailureException)
            result = Failure(
                self,
                time.time() - start_time,
                caught_exception,
                current_step,
                failure_stack_trace
            )

        self.run_special_method(self._engine.tear_down, exceptions.TearDownException)
        return result


class StoryFile(object):
    """
    YAML file containing one or more named stories, part of a collection.
    """
    def __init__(self, filename, engine, collection):
        self._filename = filename
        self._yaml = filename.bytes().decode('utf8')
        self._engine = engine
        self._collection = collection

        story_schema = {
            Optional("scenario"): Seq(Any()),
            Optional("description"): Str(),
            Optional("based on"): Str(),
        }

        if self._engine.schema.about is not None:
            for about_property, property_schema in self._engine.schema.about.items():
                story_schema[about_property] = property_schema

        story_schema['params'] = self._engine.schema.params
        story_schema['preconditions'] = self._engine.schema.preconditions

        # Load YAML into memory
        try:
            self._parsed_yaml = load(
                self._yaml,
                MapPattern(Str(), Map(story_schema))
            )
            self._updated_yaml = copy.copy(self._parsed_yaml)
        except strictyaml.YAMLError as error:
            raise exceptions.StoryYAMLError(
                filename, str(error)
            )

    def update(self, story, step, kwargs):
        self._updated_yaml[story.name]['scenario'][step.index][step.name] = \
            list(kwargs.values())[0]

    @property
    def filename(self):
        return self._filename

    @property
    def path(self):
        return Path(self._filename)

    def ordered_arbitrarily(self):
        """
        Return all of the stories in the file.
        """
        stories = []
        for name, self._parsed_yaml in self._parsed_yaml.items():
            stories.append(Story(self, str(name), self._parsed_yaml, self._engine, self._collection))
        return stories


class NewStory(object):
    def __init__(self, engine):
        self._engine = engine

    def save(self):
        """
        Write out the updated story to file.
        """
        story_file = self._engine.story.story_file
        story_file.path.write_text(
            story_file._updated_yaml.as_yaml()
        )


class StoryList(object):
    """
    A sequence of stories ready to be played in order.
    """

    def __init__(self, stories):
        for story in stories:
            assert type(story) is Story
        self._stories = stories

    def play(self):
        results = ResultList()
        for story in self._stories:
            results.append(story.play())
        return results

    def __len__(self):
        return len(self._stories)

    def __getitem__(self, index):
        return self._stories[index]



from slugify import slugify


class HitchStoryException(Exception):
    pass


class StepNotFound(HitchStoryException):
    """
    Step in story has no corresponding method in engine.
    """
    pass


class StoryYAMLError(HitchStoryException):
    """
    YAML error found parsing a story file.
    """
    def __init__(self, filename, error):
        super(HitchStoryException, self).__init__((
            "YAML Error in file '{0}':\n{1}".format(filename, error)
        ))


class StoryNotFound(HitchStoryException):
    """
    Story not found.
    """
    def __init__(self, name):
        super(HitchStoryException, self).__init__((
            "Story '{0}' not found.".format(name)
        ))


class StepException(HitchStoryException):
    """
    Exception relating to a particular step.
    """
    pass


class StepNotCallable(StepException):
    """
    The step you tried to call is not a python method.
    """
    pass


class StepContainsInvalidValidator(StepException):
    """
    Step contains a validator for which there is no corresponding argument.
    """
    pass


class WrongEngineType(HitchStoryException):
    """
    Engine should inherit from hitchstory.BaseEngine.
    """
    pass


class MoreThanOneStory(HitchStoryException):
    """
    More than one story was found matching query.
    """
    def __init__(self, stories):
        super(HitchStoryException, self).__init__(
            "More than one matching story:\n{0}\n".format(stories)
        )


class NoStories(HitchStoryException):
    """
    User tried to use .one() but no stories were found.
    """
    pass


class DuplicateStoryNames(HitchStoryException):
    """
    Two or more stories in a collection have identical or too-similar names.
    """
    def __init__(self, story1, story2):
        super(HitchStoryException, self).__init__((
            "Story '{0}' in '{1}' and '{2}' in '{3}' are identical "
            "when slugified ('{4}' and '{5}')."
        ).format(
            story1.name,
            story1.filename,
            story2.name,
            story2.filename,
            slugify(story1.name),
            slugify(story2.name)
        ))


class InvalidStoryPaths(HitchStoryException):
    """
    storypaths iterator fed to StoryCollection is invalid.
    """
    pass


class HitchStorySpecialMethodException(HitchStoryException):
    """
    Exception was raised in a special method:
    on_success, on_failure, tear_down
    """
    def __init__(self, stacktrace):
        super(HitchStorySpecialMethodException, self).__init__(
            "Stacktrace:\n{0}\n".format(stacktrace)
        )


class OnSuccessException(HitchStorySpecialMethodException):
    """
    Exception occurred in on_success method.
    """
    pass


class OnFailureException(HitchStorySpecialMethodException):
    """
    Exception occurred in on_failure method.
    """
    pass


class TearDownException(HitchStorySpecialMethodException):
    """
    Exception occurred in tear_down method.
    """
    pass



"""
User-exposed engine related code.
"""
from hitchstory import exceptions
from strictyaml import MapPattern, Any, Validator, Optional
from hitchstory.story import NewStory


def validate(**kwargs):
    """
    Decorator for validating arguments in a HitchStory step.
    """
    def decorator(step_function):
        for arg in kwargs:
            if arg not in step_function.__code__.co_varnames:
                raise exceptions.StepContainsInvalidValidator(
                    "Step {} does not contain argument '{}' listed as a validator.".format(
                        step_function.__repr__(), arg
                    )
                )
        step_function._validators = kwargs
        return step_function
    return decorator


class StorySchema(object):
    """
    Represents user-defineable parts of the hitchstory schema:

    * preconditions - properties which set up the story.
    * parameters - variables which you can feed into a story.
    * about - descriptive properties - e.g. feature names, issue ticket numbers
    """
    def __init__(self, preconditions=None, params=None, about=None):
        if preconditions is None:
            self._preconditions = MapPattern(Any(), Any())
        else:
            assert isinstance(preconditions, Validator), \
                "preconditions must be strictyaml Validator"
            self._preconditions = preconditions

        if params is None:
            self._params = MapPattern(Any(), Any())
        else:
            assert isinstance(params, Validator), \
                "params must be strictyaml Validator"
            self._params = params

        if about is not None:
            assert isinstance(about, dict), "about must be a dict of named validators"
            for key, validator in about.items():
                assert isinstance(key, str) or isinstance(key, Optional), \
                    "name must be a string or strictyaml Optional."
                assert isinstance(validator, Validator), "validator must be strictyaml Validator"

        self._about = about

    @property
    def preconditions(self):
        return self._preconditions

    @property
    def params(self):
        return self._params

    @property
    def about(self):
        return self._about


class BaseEngine(object):
    schema = StorySchema()

    @property
    def new_story(self):
        return NewStory(self)

    @property
    def preconditions(self):
        return self._preconditions

    def set_up(self):
        pass

    def on_abort(self, signal_num, stack_frame):
        pass

    def on_failure(self):
        pass

    def on_success(self):
        pass

    def tear_down(self):
        pass





def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return ''.join(x for x in text if x.isalpha() or x.isdigit() or x == "_")


def replace_parameter(thing, param_name, param):
    """
    Replace parameter name in (( and )) with value in step arguments and preconditions.
    """
    if type(thing) is str:
        if "(( {0} ))".format(param_name) in str(thing):
            return thing.replace("(( {0} ))".format(str(param_name)), str(param))
        else:
            return thing
    else:
        if type(thing) is list:
            new_thing = []

            for i, item in enumerate(thing):
                new_thing.append(replace_parameter(item, param_name, param))

            return new_thing
        elif type(thing) is dict:
            new_thing = {}

            for key, value in thing.items():
                new_thing[key] = replace_parameter(value, param_name, param)

            return new_thing
        else:
            return replace_parameter(str(thing), param_name, param)



from hitchstory.engine import BaseEngine
from hitchstory.story import StoryList
from hitchstory.story import StoryFile
from hitchstory import exceptions
from slugify import slugify
from path import Path
from copy import copy


class StoryCollection(object):
    """
    Unordered group of related stories.
    """

    def __init__(self, storypaths, engine):
        """
        Create a collection of hitch stories from a list (or generator) of
        story paths and an initialized story engine.
        """
        if not isinstance(engine, BaseEngine):
            raise exceptions.WrongEngineType(
                "Engine should inherit from hitchstory.BaseEngine."
            )
        if isinstance(storypaths, str):
            raise exceptions.InvalidStoryPaths((
                "storypaths should be a list or iterator returning a list of story files"
                " (e.g. using pathquery). Instead it was string '{0}'."
            ).format(storypaths))
        self._storypaths = storypaths
        self._engine = engine
        self._in_filename = None
        self._named = None
        self._filters = []

        self._stories = {}

        for filename in self._storypaths:
            if not Path(filename).exists():
                raise exceptions.InvalidStoryPaths(
                    "Story file name '{0}' does not exist.".format(filename)
                )
            if Path(filename).isdir():
                raise exceptions.InvalidStoryPaths(
                    "Story path '{0}' is a directory.".format(filename)
                )
            for story in StoryFile(filename, self._engine, self).ordered_arbitrarily():
                if story.slug in self._stories:
                    raise exceptions.DuplicateStoryNames(story, self._stories[story.slug])
                self._stories[story.slug] = story

    def ordered_arbitrarily(self):
        """
        Return a StoryList object containing stories filtered
        from the collection.
        """
        stories = []
        for story in self._stories.values():
            filtered = True
            for filter_func in self._filters:
                if not filter_func(story):
                    filtered = False
            if self._named is not None:
                if story.name != self._named:
                    filtered = False
            if self._in_filename is not None:
                if Path(story.filename).abspath() != Path(self._in_filename).abspath():
                    filtered = False
            if filtered:
                stories.append(story)

        # Check for non-existent inherited stories
        for story in stories:
            if "based on" in story._parsed_yaml:
                inherited_from = story._parsed_yaml['based on']
                found = False
                for search_story in stories:
                    if inherited_from == search_story.name:
                        found = True
                if not found:
                    raise exceptions.StoryNotFound(inherited_from)
        return stories

    def filter(self, filter_func):
        assert callable(filter_func)
        new_collection = copy(self)
        new_collection._filters.append(filter_func)
        return new_collection

    def in_filename(self, filename):
        assert type(filename) is str
        new_collection = copy(self)
        new_collection._in_filename = Path(filename)
        assert new_collection._in_filename.exists()
        return new_collection

    def named(self, name):
        """
        Return a single story object with the name specified.

        Only slugified names are compared. E.g. "Story NAME" and "story-name" are equivalent.
        """
        for story in self.ordered_arbitrarily():
            if slugify(name) == story.slug:
                return story
        raise exceptions.StoryNotFound(name)

    def ordered_by_name(self):
        """
        Return a list of stories ordered by name.
        """
        return StoryList(sorted(self.ordered_arbitrarily(), key=lambda story: story.name))

    def shortcut(self, *words):
        """
        Return a single story that matches all of the words.
        """
        matching = []
        for story in self.ordered_arbitrarily():
            if len([word for word in words if slugify(word) in story.slug]) == len(words):
                matching.append(story)
        if len(matching) == 0:
            raise exceptions.StoryNotFound(", ".join(words))
        elif len(matching) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join([
                    "{0} (in {1})".format(story.name, story.filename) for story in matching
                ])
            )
        else:
            return matching[0]

    def one(self):
        stories = self.ordered_arbitrarily()
        if len(stories) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join([
                    "{0} (in {1})".format(story.name, story.filename) for story in stories
                ])
            )
        elif len(stories) == 0:
            raise exceptions.NoStories()
        else:
            return stories[0]



from hitchstory import utils
from strictyaml import YAML, Any


class Arguments(object):
    """A null-argument, single argument or group of arguments of a hitchstory step."""

    def __init__(self, yaml_args, params):
        """Create arguments from dict (from yaml)."""
        self._params = params

        if yaml_args is None:
            self.is_none = True
            self.single_argument = False
        elif yaml_args.is_mapping():
            self.is_none = False
            self.single_argument = False
            self.original_args = yaml_args
        else:
            self.is_none = False
            self.single_argument = True
            self.argument = self.parameterize(yaml_args)

    def parameterize(self, value):
        for name, parameter in self._params.items():
            original_value = value.data if isinstance(value, YAML) else value
            value = utils.replace_parameter(original_value, str(name), str(parameter))
        return value

    def validate(self, validators):
        """
        Validate step using validators specified in decorators.
        """
        if not self.is_none and not self.single_argument:
            _kwargs = {}
            for key, value in self.original_args.items():
                if str(key) in validators.keys():
                    _kwargs[key] = validators[key](value._chunk).data
                else:
                    _kwargs[key] = Any()(value._chunk).data

            self.kwargs = self.parameterize(_kwargs)

    def pythonized_kwargs(self):
        pythonized_dict = {}
        for key, value in self.kwargs.items():
            pythonized_dict[utils.to_underscore_style(str(key))] = value.data \
                if isinstance(value, YAML) else value
        return pythonized_dict



from jinja2.environment import Environment
from jinja2 import FileSystemLoader
from os import path
import colorama


TEMPLATE_DIR = path.join(path.dirname(path.realpath(__file__)), "templates")


class ResultList(object):
    def __init__(self):
        self._results = []

    def append(self, result):
        self._results.append(result)

    def report(self):
        return '\n'.join([result.report() for result in self._results])


class Report(object):
    pass


class Result(object):
    def _template(self, template_name):
        env = Environment()
        env.loader = FileSystemLoader(TEMPLATE_DIR)
        return env.get_template(
            path.basename(path.join(TEMPLATE_DIR, template_name))
        ).render(
            result=self,
            Fore=colorama.Fore,
            Back=colorama.Back,
            Style=colorama.Style,
        )

    @property
    def exit_code(self):
        return 0

    @property
    def duration(self):
        return self._duration

    @property
    def story(self):
        return self._story


class Success(Result):
    def __init__(self, story, duration):
        self._story = story
        self._duration = duration

    def report(self):
        return self._template("success.jinja2")


class FailureException(object):
    def __init__(self, exception):
        self._exception = exception

    @property
    def obj(self):
        return self._exception

    @property
    def obj_type(self):
        return str(self._exception.__class__.__name__)

    @property
    def docstring(self):
        return str(self._exception.__doc__)

    @property
    def text(self):
        return str(self._exception)


class Failure(Result):
    def __init__(self, story, duration, exception, failing_step, stacktrace):
        assert type(exception) is Exception or RuntimeError
        self._story = story
        self._duration = duration
        self._exception = FailureException(exception)
        self._failing_step = failing_step
        self._stacktrace = stacktrace

    @property
    def exit_code(self):
        return 1

    @property
    def exception(self):
        return self._exception

    @property
    def stacktrace(self):
        return self._stacktrace

    @property
    def story_failure_snippet(self):
        """
        Snippet of YAML highlighting the failing line.
        """
        snippet = "{before}\n{bright}{lines}{normal}\n{after}".format(
            before=self._failing_step.yaml.lines_before(2),
            lines=self._failing_step.yaml.lines(),
            after=self._failing_step.yaml.lines_after(2),
            bright=colorama.Style.BRIGHT,
            normal=colorama.Style.NORMAL,
        )
        indented_snippet = "    " + snippet.replace("\n", "\n    ")
        return indented_snippet

    def to_dict(self):
        return {
            'story': self.story.to_dict(),
            'step': self.step.to_dict() if self.step else None,
            'exception': str(self.exception),
            'exception_type': "{}.{}".format(
                type(self.exception).__module__, type(self.exception).__name__
            ),
        }

    def report(self):
        return self._template("failure.jinja2")
